/* (c) Joerg Schulenburg 2000-2019 GPL

 this is part of spin-package

 it is for division of matrix multiplication 
 in parallelizable block matrix multiplication (see block.h)

 */
#include "block.h"
#include <unistd.h> /* usleep() */
#include <stdio.h>  /* printf */

#ifdef CONFIG_PTHREAD   /* ToDo: movto mthreads.h, or write new OpenMP-code */
#ifndef _OPENMP /* OPENMP is replacing PTHREAD */
 /* pt_attr: SunOS has SCOPE_PROCESS as default, we need SCOPE_SYSTEM */
 pthread_attr_t pt_attr;
 pthread_cond_t pt_cond1;  /* conditions master -> childs */
 pthread_cond_t pt_cond2;  /* conditions childs -> master */
 pthread_mutex_t pt_mutex1; /* mutex */
 pthread_mutex_t pt_mutex2; /* mutex */
 pthread_mutex_t token_ring[B_NUM*4]; /* 4 rings of double(quad) length */
 pthread_barrier_t pt_barrier; /* 2017-01-18 for new better mixed code (MPI+PT) */
#else /* OPENMP-replacements? PTHREAD-dummys! */
 volatile int omp_token=0;
 volatile int omp_token2=0;
#endif /* ~_OPENMP */
#if defined(__APPLE__)   /* minimum implementation by A.Armea 2013, 2018-11 */
/* see block.h */
int pthread_barrier_init(pthread_barrier_t *barrier,
    const pthread_barrierattr_t *attr, unsigned int count) {
    if(count == 0) { errno = EINVAL; return -1; }
    if(pthread_mutex_init(&barrier->mutex, 0) < 0) { return -1; }
    if(pthread_cond_init(&barrier->cond, 0) < 0) {
        pthread_mutex_destroy(&barrier->mutex);
        return -1;  }
    barrier->tripCount = count;
    barrier->count = 0;
    return 0;
}
int pthread_barrier_destroy(pthread_barrier_t *barrier) {
    pthread_cond_destroy(&barrier->cond);
    pthread_mutex_destroy(&barrier->mutex);
    return 0;
}
int pthread_barrier_wait(pthread_barrier_t *barrier) {
    pthread_mutex_lock(&barrier->mutex);
    ++(barrier->count);
    if(barrier->count >= barrier->tripCount) {
        barrier->count = 0;
        pthread_cond_broadcast(&barrier->cond);
        pthread_mutex_unlock(&barrier->mutex);
        return 1; }
    else {
        pthread_cond_wait(&barrier->cond, &(barrier->mutex));
        pthread_mutex_unlock(&barrier->mutex);
        return 0;
    }
}
#endif // __APPLE__
#endif // CONFIG_PTHREAD

int mpi_n=1, mpi_i=0;   /* number of MPI nodes, no. of this node */
int pt_n=1;             /* number of threads per node 1..B_NUM */
int mpi_n_bits=1, pt_n_bits=1; /* bits needed to store mpi_i and pt_i, MPI */
/* ToDo: only 32 bit? -> ULong */
/* see b_ini: node_ofs=mpi_ofs[mpi_i]; b_ofs[i]=0..node_n1 */
unsigned long node_ofs, node_n1; /* ofset and length of subspace this node */
unsigned long b_ofs[B_NUM+1];	      /* see block.h */
unsigned long mpi_ofs[MPI_MAX_NUM+1]; /* see block.h */
unsigned long max_b_len;        /* maximum b_len to ensure mpi balance */
/* ToDo: move b_hnz to h_hnz h_file.c!? where h_ is defined */
unsigned long b_hnz[B_NUM];	/* block NonZero H-entries */
unsigned long b_snz[B_NUM];	/* stored hnz block entries */
unsigned long b_sn1[B_NUM];	/* stored hnz block lines */
unsigned long len_full_block;   /* number of elements per block */

/* 2019-01 move struct thxy b_hbuf to hilbert.h thbuf = struct shbuf 
 *  ... reason: it was to big for B_NUM=32 and NAH=AHEAD=2048 NZX=196 */
// struct thxy b_hbuf[B_NUM];      /* buffer for faster r/w-access */ 
int         b_bnz[B_NUM];       /* num valid nz-data in buffer hbuf */  
int         b_bn1[B_NUM];       /* num valid n1-lines in buffer hbuf */  

// moved from spins.c to block.c multithread functions:
/* local variables for pthreads (tasks), communication to/from tasks */
//typedef struct {
//  int num;     /* to identify itself (ro) */
//  void *data;  /* to send data-block (set by thread 0, read by workers) */
//  int  task;   /* master set 1 to run task + waits for 0, task sets 0 @end */
//  void (*worker)(void *data); /* parallel working function */
//  /* mutex for each thread? */
//} main_pts;
volatile main_pts main_pt[B_NUM]; /* [un]locked by mutex1 */

#ifdef CONFIG_PTHREAD
/* send void-func + data pointer to waiting pthread and restart it */
/*  ... usually this is done via loop over pt_n to let pthread_barrier work */
/* 2016-10 g++-6.2 + pthread + blas segfaults for stacksize 8K (ulimit -s) */
void pt_start( int thread_i, void (*worker)(void *data), void *data ){
    pthread_mutex_lock(&pt_mutex1);
    // printf("DBG: send work to task %d\n", thread_i);
#pragma omp critical
{   /* volatile data */
    main_pt[thread_i].data=data;
    main_pt[thread_i].worker=(void (*)(void *))worker;
    main_pt[thread_i].task=1; /* signal worker thread i to run (last!) */
}
    /* ToDo: do next only once after storing all tasks
             or send signal only to the job */
    pthread_cond_broadcast(&pt_cond1); /* data send to threads */
    pthread_mutex_unlock(&pt_mutex1); /* redefined as dummys if _OPENMP */
#ifdef _OPENMP
#pragma omp flush (main_pt)
#endif
}

/* waiting for threads finishing there task given by pt_start */
void pt_end_wait( int thread_i ){ // Jul22 rename from parallel_wait
#ifndef _OPENMP
  pthread_mutex_lock(&pt_mutex1);
  while ( main_pt[thread_i].task )  /* volatile! */
    pthread_cond_wait(&pt_cond2,&pt_mutex1); /* unlock+wait_for_child+lock */
  pthread_mutex_unlock(&pt_mutex1);
#else
  while ( main_pt[thread_i].task )  /* volatile! */
    usleep(10000); /* 10ms ToDo  */
#endif
}

// PT_Barrier + rc = (int)1 for one thread, rest get 0
//  needs initial pthread_barrier_init()
//  is needed by MPI + PT mixed code where only thread 0 does MPI
int pt_barrier_wait() {  // better name pt_barrier_wait()
  int rc= pthread_barrier_wait(&pt_barrier); /* 2017-01-18 wait all threads */
  return rc;  // pt_barrier defined in block.c + .h
}
#else /* ~CFG_PTHREAD dummy parallel functions (serialized pthreads) */
void pt_start( int thread_i, void (*worker)(void *data), void *data ){
  worker(data); }
void pt_end_wait( int thread_i ){}
int pt_barrier_wait() { return 1 };
#endif /* CFG_PTHREAD */



/* release mem to satisfy memleak checker */
void b_free(){
   /* free dynamic arrays */
   /* nothing to do v2.36 */
}

/* initialization of thread blocks for n1/mpi_n elements on each node,
 * n1 is whole matrix size, called by every node, only thread0
 */
int b_ini(unsigned long n1,int vvv){
  int i, num_full_blocks;
  unsigned long len_rest_block, ofs, len;
  
  max_b_len=0;
          
  /* pt_n is numthreads per node */
  for (i=0;i<=      B_NUM;i++)   b_ofs[i]=0; /* initialize array 0..pt_n */
  for (i=0;i<=MPI_MAX_NUM;i++) mpi_ofs[i]=0; /* initialize array 0..mpi_n */
  for (i=0;i<B_NUM;i++) {  /* initialize all arrays (1..pt_n..B_NUM)-1 */
    b_sn1[i]=0; /* belongs to h_file!? */
    b_hnz[i]=0;
    b_snz[i]=0;
    b_bnz[i]=0;
    b_bn1[i]=0;
  }
  /* block length depends from number of mpi nodes and threads per node */
  len_full_block=(n1+pt_n*mpi_n-1)/(pt_n*mpi_n); /* #loops/thread 0..n1 */
#if defined(HAVE_MMAP) && defined(CFG_PAGESIZE)
  /* mmap offset must be a multiple of PAGESIZE! fill up block length */
  if (len_full_block>0)
      len_full_block=((len_full_block-1)/CFG_PAGESIZE+1)*CFG_PAGESIZE;
#endif /* CFG_PAGESIZE */

  MPI_Barrier(MPI_COMM_WORLD); /* debug */

  /* calculate num of full size blocks and length rest block */
  num_full_blocks= n1 / len_full_block;     /* range: 0..(pt_n*mpi_n-1) */
  len_rest_block=  n1 - len_full_block * num_full_blocks;  /* 0..(n1-1) */

  if (!mpi_i)
    printf("DBG.b_ini: mpi*pt*thread_n1= %3d * %2d *   %10lu = %10lu >=n1\n",
       mpi_n, pt_n, len_full_block, mpi_n*pt_n*len_full_block);
  if (pt_n>1 && !mpi_i)
    printf(" pt_blks: num*length+rest= %4d * %7lu + %8lu = %10lu =n1\n",
       num_full_blocks, len_full_block, len_rest_block, n1);
  if (mpi_n>1 && !mpi_i)
    printf("mpi_blks: num*length+rest= %3d * %8lu + %8lu = %10lu =n1\n",
       num_full_blocks/pt_n, pt_n*len_full_block,
       n1 - pt_n*len_full_block*(num_full_blocks/pt_n), n1);

  node_ofs=0; node_n1=0;
  MPI_Barrier(MPI_COMM_WORLD); /* debug */

  /* setup mpi blocks (offsets and implicite lengths) */
  ofs=0; len=n1; max_b_len=len_full_block*pt_n;
  for (i=0; i<mpi_n; i++) {
    mpi_ofs[i  ]=ofs;
    mpi_ofs[i+1]=ofs+((len<max_b_len)?len:max_b_len); /* next-ofs */
    ofs =mpi_ofs[i+1];
    len-=mpi_ofs[i+1]-mpi_ofs[i];
    /* only mpi_i==0 prints out table for other mpi-tasks */
    if ((!(vvv&4)) || mpi_i) continue; /* dont print */
    if ((i==2 && mpi_n>4)) printf("DBG_MPI: ...\n");
    if ((i< 2 || i>=mpi_n-2)) /* ToDo: better if((i&(i-1))==0||i+1==mpi_n) */
      printf("DBG_MPI: node= %4d     mpi_len= %8lu mpi_ofs= %10lu   %10lu\n",
              i, mpi_ofs[i+1]-mpi_ofs[i], mpi_ofs[i], mpi_ofs[i+1]);
  }

  node_n1 =mpi_ofs[mpi_i+1]
          -mpi_ofs[mpi_i];  /* length of block in this node */
  node_ofs=mpi_ofs[mpi_i];  /* node offset, do we need it? */

  /* setup pthread blocks (offsets and implicite lengths) */
  if ((vvv&2) && mpi_i) usleep((mpi_i%8)*10000);
  ofs=0; len=node_n1; max_b_len=len_full_block;
  for (i=0; i<pt_n; i++) {
    b_ofs[i  ]=ofs;
    b_ofs[i+1]=ofs+((len<max_b_len)?len:max_b_len);
    ofs =b_ofs[i+1];
    len-=b_ofs[i+1]-b_ofs[i];
    /* print mpi_i==0 only, hybrid code removed later for simplicity? */
    if ((!(vvv&4)) || (0<mpi_i && mpi_i<mpi_n-1)
                   || (1<i     &&     i< pt_n-2) || pt_n==1
                  /* || b_ofs[i+1]==b_ofs[i] */) continue; /* dont print */
    printf("DBG_PT:  node=%3d blk=%3d b_len= %8lu   b_ofs= %10lu   %10lu\n",
              mpi_i, i, b_ofs[i+1]-b_ofs[i], b_ofs[i], b_ofs[i+1]);
  }

  return 0;
}
