/*
  what it does:
   measure speed of mpi-transfer between tasks (more than just 2) and
   switching source and destination of transfer at every cycle (thats why
   it is stressing the network) or simply do MPI_alltoall
  problems:
    unknown dynamic need of memory, 2tasks/node*16threads+affinity(+HT)
    strace: access("/sys/class/infiniband", F_OK)   = 0  2018.05
    strace: /sys/devices/system/node/node0/meminfo # 128GB
            /sys/devices/system/node/node1/meminfo # 128GB
            sched_getaffinity
            /sys/devices/system/cpu
            /sys/devices/system/cpu/online # 0-31
            mmap+mprotect+mbind+clone ... sched_yield() 90% ...
            ~/.openmpi/components /dev/shm/... /dev/kcopy/0..255
            SPIN ... futex ... 2nd-rcfg = hang 2018-05

  linux:
    mpicc -O2 -lm -o mpi_stress mpi_stress.c
    mpicc -O2 -lm -D CFG_CPUSET=0x00ff -o mpi_stress mpi_stress.c
    # cpuset is a bit pattern (hex) for available cpu cores per node
    #   if you want bind tasks to cores using sched_setaffinity()
    mpirun -np 8 ./mpi_stress --bsize 26 | tee mpi_stress.log
    # example with tcp + tcp_over_ib
    mpiexec --mca btl self,sm,tcp --mca btl_tcp_if_include eth0 --mca
     btl_tcp_if_include ib0 --hostfile $PBS_NODEFILE --bynode -np 8 numactl
     --localalloc --physcpubind 0 ./mpi_stress --bsize 30 --alltoall 
# comp2: QDR-IB + ompi-1.8.4 2015-02
#  --mca btl ^openib --mca btl_tcp_if_include eth0 2*1 132us  1.1e7*1 =11MB/s M
#  --mca btl ^openib --mca btl_tcp_if_include eth0 2*16 81us  1.4e6*16=22MB/s M
#  --mca btl ^openib --mca btl_tcp_if_include ib0  2*1  46us  5.0e8*1 =0.5GB/s 
#  --mca btl ^openib --mca btl_tcp_if_include ib0  2*16 62us  1.6e8*16=2.6GB/s
#  tcp-over-ib:  60% BW + 160kB ppLatenz (MTU=64KB)
#  --mca btl ^tcp    --mca btl_tcp_if_include ib0  2*1  1.6us 2.7e9   =2.7GB/s
#  --mca btl ^tcp    --mca btl_tcp_if_include ib0  2*16 7.2us 2.7e8*16=4.3GB/s
#  --bind-to core -report-bindings
# IB-scaling 2*1,3*1,4*1,8*1,16*1 = 2.7GB/s,  SHM=....? estimation 2*16?

  check raw networkspeed first (2 nodes)
   
  see mpi_stress.gpl and at bottom of this file for some data!
  
*/
const char *js_version_string="v2021-11-20 (timing: v2021-11-20)";
/* Changelog:
    2021-11-20 change to pairwise to find bad connections, min/max timing
      test gcc-8.4.1 on hpc21 
    2015-03-31 add minfree test (problems on 5700 cores and mpich, 130MB free) 
    2015-02-23 - add --ckpthelp[=..] for checkpoint/restart testing (dmtcp)
                 removed 2021-11
    2014-07-27 - add --minfree=.. to check vmem needed for mpi-calls
    2014-07-01 - cpu-pinning per option
    2014-06-19 - add option --bsize0 log2_min_block_size (debugging)
    2014-06-16 - add min loop check (loops >= (mpi_n-1)) for accuracy
    2013-06-21 - more compact output, aBW=SumBW (a=aggregated), t/s=time[s]
    2011-10-24 - better timing error for A2A too + memory constraints
    2011-10-20 - limit msg size to 1GB (MPI restriction for MPI_CHAR)
               - blocksize from min to max (old: max to min)
               - compute error by time variance and adapt runtime
               - check for bit errors
    2008-08-23 - change MPI_Sendrecv loop to avoid bad measurement on NUMA
                 systems (Altix4700: 8-9GB/s measured, but 1.3GB/s real) 
    2008-08-20 - fix race condition, have lead
                 to speed-down-jumps, MPI_ERR_TRUNCATE and dead-locks at end
                 (thanks to David Gingold)
               - better option handling
  Problems:
    - one loop is only part of a full all2all, where distribution
      of remote and local connection and speed results depend from
      the distribution, getting good results after (threads/node-1)
      cycles but that would be takes long for big Nodes! (2011-10)
      ToDo: successiv searching slowest and fastest N/2-pairing ???
        i searches i+1..i-1 for slowest connect, after that next
        unpaired node does the same until all nodes paired (N steps)
        2nd run same with fastest pairing (all local! not useful?)

  ToDo:
    - mehrere paralele transfers auf Nodes aber kein intra-node transfer
      between nodes only (not inside), better hiding latencies
      get better transferrates for smaller packages (if not synchron)
    - alloc asize * bsize block z.B. 2048 * 1MB = 2GB/task
      then transfer node[i  ].mem1[block*(j%asize)]
                 to node[i+j].mem2[block*(j%asize)]
            more realistic
      old was: transfer node[i  ].mem1[0] to node[i+j].mem2[0]
           arteficial full mem transfers
      better: v1 += sparseH * v2 # using fixed nz/line
    - async mpi_stress + cpu for max BW to test old elkos of DC-DC converter
      s. mkautz 2014-06
    - CPU_SET like memspeed.c (break 32 cores limit, or slurm cpu pinning)
    - use local mem + transfer mem (do async memset|memcopy or similar
      to avoid thread jumping in SMPs, which generate arteficial
      transfer speeds)  do we understand that?
    - old timings?
    - send more random bit pattern (checks the memory correctness)
    - add initial 2 node latency and speed (able to compute theoretical max)
      compute min vs. max of pair speed (detect local + remote pairs)
    - SMP influence of cache coherence (disable?)
    - add a all send to one OOM-pressure test (s. vector.c mpich failes)
    - check 0-to-i homogenity add beginning (net topology)

  FAQ:
   - Q: Why not MPI_alltoall? Could better optimize for topology.
     A: Must have much (N times) more memory for buffering transmitted data.
        In real app we can transfer parts, compute and free the buffer.

*/

#define _GNU_SOURCE 1 // this defines __USE_GNU in features.h for sched.h

#include <stdlib.h>
#include <stdio.h>
#include <mpi.h> 
// #include <time.h>
#include <malloc.h> /* memalign() */
#include <string.h>
#include <math.h>  /* fabs() */

#include <unistd.h>  /* declare sleep() */

#ifdef CFG_CPUSET // switch on if sched_setaffinity will benefit
# include <sched.h>   // declare sched_setaffinity()
# ifndef _FEATURES_H  
#  warning "_FEATURES_H is not defined, try top add #include <features.h>"
# endif
# ifndef __USE_GNU                                                 
#  warning "expect __USE_GNU defined, but it isn't, try -D_GNU_SOURCE=1"
# endif
#endif

int mpi_n, mpi_i,
    rankshift=1; /* destination rank mpi_i+rankshift 2021-11 */
double mintime=0.5; /* minimum seconds run time (max. is 2*mintime) */
double * tarray = NULL; /* Allgather run times */
char *buf1, *buf2; /* transfer buffers */
static int header=0;

int test_sendrecv(unsigned long len){
  unsigned long i1, i2, i3;
  int i5, rank2, itmax, itmin, iavg1=0, iavg2=0;
  double t1, t2, tmax, tmin, tavg, tavg1, tavg2;  /* run time */
  MPI_Status status;
  if (!header) { header=1;
    if (!mpi_i)
    printf("# tasks msgsize/B lg2loops t/s"  /* BW=bandwidth per wire */
           " latency/s BW[B/s]  MinMaxBW,idx                   nfast,nslow\n");
           
  }
  i3=1;  // log2 of minloops
  // ToDo: dont double on timing errors, just repeat and print min/max time
  //   this would be better on switches which perform poor at critical usage
  //  t>mintime abbrechen (done), min/maxspeed ab mintime/2^4 ???
  // ToDo: 2014-06 better i1=(mpi_n-1)<<i3 for full cycle?
  //      bei kautz = 5700 (2^12..2^12) sehr viel!? ToDo
// add -DCPU_LOAD for C/R tests where pure mpi-communication fails 2015-02
#ifdef CPU_LOAD  /* helps dmtcp checkpointing outside TCP-calls */
  double x=17; int i4; static int xf=1;
  if (xf) { xf=0;
      t2=0.0; t1=MPI_Wtime();
      for (i4=0;i4<100000000;i4++) x=0.5*(x+17/x); /* sqrt(x) ca. 1s/1e8 */
      t2=MPI_Wtime()-t1; 
      if(!mpi_i) printf("# CPU_LOAD= %.3f s per loop\n", t2); /* vs. t_MPI */
  }
#endif
  rank2 = (mpi_i + ((((mpi_i/abs(rankshift))&1)?-1:1)*rankshift) + mpi_n) % mpi_n; /* +1...n/2 */
  // if (mpi_i/(2*rankshift) >= mpi_n/(2*rankshift)) rank2=mpi_i; /* self */
  // if (mpi_n<9) printf("rank i= %d to j= %d\n", mpi_i, rank2);
  for (i1=1l<<i3;i1/*stop at int overflow*/;i1<<=1,i3++){
    // i1 = log2 of next inner loops before next time measurement
    // i2 = loops from start time t1
    t2=0.0; t1=MPI_Wtime();  // 2021-11 inside loop, because Allgather-sync
    for (i2=0;2*i2<i1;i2++) {  /* include shorter runs to spare usertime */
      /* sendcount is integer 32bit, max block is 2GB! */
#ifdef CPU_LOAD /* warm up? should be in 2nd thread */
      for (i4=0;i4<100000000;i4++) x=0.5*(x+17/x); /* sqrt(x) ca. 1s/1e8 */
#endif
      MPI_Sendrecv(buf1,len,MPI_BYTE,rank2,0, /* send, rcv ... */
                   buf2,len,MPI_BYTE,rank2,0,MPI_COMM_WORLD,&status);
      /* do a 2nd call (pong) to avoid wrong measurement on NUMA systems */
      MPI_Sendrecv(buf2,len,MPI_BYTE,rank2,0,
                   buf1,len,MPI_BYTE,rank2,0,MPI_COMM_WORLD,&status);
    }
    t2=MPI_Wtime()-t1;
    /* avoid a race condition, added time will vanish for bigger i3 */
    // ToDo: replace by min/max to get rankshift+pairwise SMP/TCP ranges
    // Bcast = broadcast sends data from rank 0 to rank 1,2,3,...
    // MPI_Bcast(&t2, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD); /* t.rank0 only */
    MPI_Allgather(&t2, 1, MPI_DOUBLE, tarray, 1, MPI_DOUBLE, MPI_COMM_WORLD);
    // here all tasks synced by Allgather to slowest
    t2 = tarray[0]; tmin=tmax=t2; tavg=tavg1=tavg2=0.0; 
    itmax=itmin=iavg1=iavg2=0;
    for (i5=0;i5<mpi_n;i5++) {
      if (tmax<tarray[i5]) { tmax=tarray[i5]; itmax=i5; }
      if (tmin>tarray[i5]) { tmin=tarray[i5]; itmin=i5; }
      tavg += tarray[i5];
    }
    for (i5=0;i5<mpi_n;i5++) {
      /* compute lower and upper average, dipol moments */
      if (tarray[i5]<=(tmin+tmax)/2) { tavg1 += tarray[i5]; iavg1++; }
      else                           { tavg2 += tarray[i5]; iavg2++; }
    }
    tavg /= (1.0*mpi_n);
    // t3+=MPI_Wtime()-t1-t2;
    /* t2 must be the same on all threads! */
    if (t2>mintime) break;
    // ToDo: ... < (2*t3+t2)/2 see comp2 256 tasks
    if (t2>10.0 && !mpi_i) // some alive messages
    printf("# %4d %9lu %6lu %6.2f %8.2e  %8.2e MiMa %8.2e %3d %8.2e %3d\n",
      mpi_n, len, i3, t2, t2/i1, i1*1.*len/tavg,
       i1*1.*len/tmax, itmax, i1*1.*len/tmin, itmin);
  }
  if (!mpi_i)
    //        N   size loops time t/cyc  BW/wire  MinMaxBW,idx     nfast,nslow
    printf("  %4d %9lu %6lu %6.2f %8.2e  %8.2e MiMa %8.2e %3d %8.2e %3d %3d %3d\n",
      mpi_n, len, i3, t2, t2/i1, i1*1.*len/tavg,
       i1*1.*len/tmax, itmax, i1*1.*len/tmin, itmin, iavg1, iavg2);
  return 0;
}

int test_alltoall(unsigned long len){
  unsigned long i1, i2, i3;
  double t1, t2, t3=0;
  if (!header) { header=1;
    if (!mpi_i)
    printf("# tasks msgsize/B lg2loops t/s"
           " latency/s aBW[B/s] BW[B/s] Err%%\n"); /* BW=bandwidth*/
  }
  i3=0;
  t1=MPI_Wtime();
  for (i2=0,i1=1l<<i3;i1;i1<<=1,i3++){
    for (;i2<i1;i2++) {  /* include shorter runs to spare usertime */
      MPI_Alltoall(buf1,len,MPI_BYTE,
                   buf2,len,MPI_BYTE,MPI_COMM_WORLD);
      // i=1..n: MPI_Send(buff+i*len*sizeof(MPI_BYTE),len,MPI_BYTE,i,...      
      // i=1..n: MPI_Recv(buff+i*len*sizeof(MPI_BYTE),len,MPI_BYTE,i,...      
    }
    t3=t2; t2=MPI_Wtime()-t1;
    /* avoid a race condition, added time will vanish for bigger i3 */
    MPI_Bcast(&t2, 1, MPI_DOUBLE, 0, MPI_COMM_WORLD);
    // t3+=MPI_Wtime()-t1-t2;
// on bad networks compute time may increase dramatically 2013-10-18
//    if ((i2>1 && t2 > 64*mintime) ||   /* hard end, saw problems on tcp GbE at 256B */
//      (fabs(2*t3-t2)*200<(t2+2*t3) && t2>mintime) ) break; /* 0.5% */
    if (t2>mintime) break;
    if (t2>10 && !mpi_i)
    printf("# %4d %9lu %6lu %6.2f %8.2e  %8.2e %8.2e %5.2f\n",
      mpi_n, len, i3, t2, t2/i1, i1*1.*len*mpi_n*mpi_n/t2, 
      i1*1.*len*mpi_n/t2, (2*t3-t2)*200/(t2+2*t3));
      /* time-error:  2*oldrun(oldloops) vs. 1*newrun(2*oldloops) */
  }
  /* we show msgsize as single transfer message bow, 1.. */
  if (!mpi_i)
    printf("  %4d %9lu %6lu %6.2f %8.2e  %8.2e %8.2e %5.2f\n",
      mpi_n, len, i3, t2, t2/i1, i1*1.*len*mpi_n*mpi_n/t2,
       i1*1.*len*mpi_n/t2, (2*t3-t2)*200/(t2+2*t3));
  return 0;
}

/* HOWTO get minfree of nodes by script?
 * srun -s -N 12 bash -c "(grep MemFree /proc/meminfo;uname -n)|xargs"|\
  awk 'BEGIN{m0=9e9;m1=0;m2=0;m3=0;}\
    {m4=$2/1024.;if(m4<m0){m0=m4;m5=$4};m3=((m3<m4)?m4:m3);m1+=m4;m2++;}\
    END{print "min/max/sumMB",m0,m3,m1,m2,m5}'  # 0.7s
 * for x in $(seq 972);do n=scx-m$((x/27))n$((x%27));
 *  ssh -o ConnectTimeout=5 $n "(grep MemFree /proc/meminfo;uname -n)|xargs";done # ...
 * fast log2 ssh tree?  
 *
 */
int get_minfree(){  /* 2015-03-31 compute minimum free mem by trying alloc */
/* problem is having multiple threads sharing same (small) memory
   and wont crash if using mpi for synchronization, where mpi needs mem too
   sync by sleep will be slow (21s)
   there would be no problem if tasks/node==1
   for tasks/node>1  there could also be a race condition
   write global shared randomnum to /dev/shm to detect tasks/node>1 ?
     or read SLURM_NTASKS_PER_NODE if used
 */
  char *buf[32]; /* max. 4GB */
  int i0= 31, i;
  long j, mem_eaten=0, max_eaten=0; /* 32bit */
  if (sizeof(long)==sizeof(size_t) && sizeof(size_t)>=64) i0=63;
  /* eating the rest of memory */
  for (i=i0;i>=10;i--) {  /* 2G 1G 512M 256M ... 1kB */
    buf[i]=(char *)malloc(((size_t)1)<<i); /* needs minimum 32bit int */
    if (buf[i]) { 
      mem_eaten |= ((size_t)1)<<i;
      for (j=0;j<1<<(i-10);j++) buf[i][j<<10]=j&255; /* touching mem */
    }
    // if((i&3)==0)sleep(1); // 31s for syncing (very bad solution)
  }
  for (i=i0;i>=10;i--) if (buf[i]) free(buf[i]);
//  if ((mpi_i&(mpi_i-1))==0) /* ToDo: get min/max of all threads */
//    printf("# task%03d mem eaten kB: %7lu\n", mpi_i, mem_eaten>>10);
  j=mem_eaten;
  MPI_Allreduce(&j, &mem_eaten, 1, MPI_LONG, MPI_MIN, MPI_COMM_WORLD);
  MPI_Allreduce(&j, &max_eaten, 1, MPI_LONG, MPI_MAX, MPI_COMM_WORLD);
  if ((mem_eaten>>10)+1==1<<i0 && !mpi_i)
    printf("WARNING: memeater failed, please ulimit vmem to max. 2GB\n");
  if(!mpi_i) printf("MinMaxFreeKB/Task: %7.0f %7.0f\n", mem_eaten/1024.,
                                                        max_eaten/1024.);
  return 0;
} /* minfree set */


int main(int argn, char *argv[]) {
  unsigned long i1,
      minfree_kB=~0L, /* 2014-07-27 memeater reserve to test mpi-calls */
      bsize0=1, /* 2014-06-19 for debugging start with 1...bsize */
      bsize;  /* buffer size or message size, 2^bsize2 */
  int version, subversion, i, i2,
      pin_flag=0, /* thread to processor pinning 2014-07-01 */
      num_bflips=0, /* 2014-06 count byteflip-errors */
      bsize2=26, /* log2 of bsize = 64MB */
      pairs=1,  /* rings (src!=dst, default) versus pairs (src==dst) */
      ata=0,    /* MPI_Sendrecv (default) versus MPI_Alltoall */
      align=1;  /* align memory to 64 (default) or not */
  char *s1;
  setvbuf(stdout,NULL,_IONBF,0);
  if (MPI_Init(&argn, &argv) != MPI_SUCCESS) exit(1);
  MPI_Comm_size(MPI_COMM_WORLD, &mpi_n);
  MPI_Comm_rank(MPI_COMM_WORLD, &mpi_i);
  MPI_Get_version(&version, &subversion);
  if (!mpi_i)
    printf("# mpi_stress %s\n", js_version_string);
    // debug info stdlib.getenv()
  if (!mpi_i) {
    s1="PBS_NUM_NODES"; 
    if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
    s1="PBS_NUM_PPN";  /* processes per node */
    if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
    s1="SLURM_CLUSTER_NAME";
    if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
    s1="SLURM_NODELIST";
    if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
    s1="SLURM_NNODES";
    if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
    s1="SLURM_NTASKS_PER_NODE";
    if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
    s1="SLURM_TASKS_PER_NODE";
    if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
    s1="OMPI_COMM_WORLD_SIZE"; /* NODES */
    if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
    s1="OMPI_COMM_WORLD_LOCAL_SIZE"; /* PPN processes per node */
    if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
    /* OMPI: Aggressive or Degraded performance mode */
    /* default: degraded on oversubscribe mode */
    s1="OMPI_MCA_mpi_yield_when_idle";
    if (getenv(s1)) printf("MPI: %s=%s\n",s1,getenv(s1));
  }
  for (i=1;i<argn;i++) {
    if (i+1<argn && !strcmp(argv[i],"--bsize0")) { /* start blocksize */
            i++; bsize0=1l<<atol(argv[i]); continue; }
    if (i+1<argn && !strcmp(argv[i],"--bsize")) {
            i++; bsize2=atol(argv[i]); continue; }
    if (i  <argn && !strncmp(argv[i],"--bsize0=",9)) { /* 2014-07 add opt= */
            bsize0=1l<<atol(argv[i]+9); continue; }
    if (i  <argn && !strncmp(argv[i],"--bsize=",8)) {
            bsize2=atol(argv[i]+8); continue; }
    if (i  <argn && !strncmp(argv[i],"--mintime=",10)) {
            mintime=atof(argv[i]+10); continue; }
    if (i+1<argn && !strcmp(argv[i],"--mintime")) {
            i++; mintime =atoi(argv[i]); continue; }
    if (i  <argn && !strncmp(argv[i],"--minfree=",10)) {
            minfree_kB =atoi(argv[i]+10); continue; }
    if (i  <argn && !strncmp(argv[i],"--rankshift=",12)) {
            rankshift =atoi(argv[i]+12); continue; }
    if (!strcmp(argv[i],"--alltoall")) { ata=1;   continue; }
    if (!strcmp(argv[i],"--pairwise")) { pairs=1; continue; }
    if (!strcmp(argv[i],"--noalign"))  { align=0; continue; }
    /* pin_* = processor core/thread pinning */
    if (strncmp(argv[i],"--pin_ncpu=",11)==0) { pin_flag=1; continue; }
    if (strncmp(argv[i],"--pin_ofs=", 10)==0) { pin_flag=1; continue; }
    if (strncmp(argv[i],"--pin_dist=",11)==0) { pin_flag=1; continue; }
    if (strncmp(argv[i],"--pin_size=",11)==0) { pin_flag=1; continue; }
    if (!mpi_i) { printf(
      "valid options: [--bsize0=<log2minsize>] [--bsize=<log2maxsize>]\n"
      "  [--mintime=<f32seconds>] [--minfree=<kB_left_by_memeater>]\n"
      "  [--alltoall|{[--pairwise] [--rankshift=<1..n/2>]}] [--noalign]\n"
      "  [--pin_ncpu=16 --pin_ofs=0 --pin_dist=1 --pin_size=1]\n"
      "\n"
     ); exit(1); }
  }
  /* Warning added JS2011-10, msgsize is int32 */
  if ((ata==0 && bsize2>30)
   || (ata==1 && 1l<<bsize2 > mpi_n*1l<<30)) {
    if(!mpi_i)
      printf("# WARNING: MPI can handle 2^30 Byte-blocks (1GB) at maximum\n");
    /* ToDo: memory size and blocks differ in case of alltoall, mem/task? */
    if (ata==0) bsize2=30;
    else for ( ; 1l<<bsize2 > mpi_n*1l<<30; bsize2--); // set to maximum
  }
  bsize = 1l << bsize2;
  if (bsize0<    1) bsize0=1;
  if (bsize0>bsize) bsize0=bsize;
  tarray = (double *) malloc ( mpi_n * sizeof(double) );
  if (!tarray) { printf("ERR: malloc failed\n"); MPI_Abort(MPI_COMM_WORLD,3); }

#if !defined(CFG_CPUSET)
  if (pin_flag==1 && !mpi_i)
    printf("WARNING: mpi_stress compiled without CFG_CPUSET\n");
#endif
#if defined(CFG_CPUSET) && defined(__USE_GNU)
  if (pin_flag==1) { 
   /* processor/thread pinning, ToDo: use ompi-capabilities! */
   int j1;
   // ToDo: openmp code missing
   // this processor bounding is useful for the case, that the jobsystem
   //   does not do this automatically
   // CFG_CPUSET is the bitmap for available SMP-CPUs per node
   // CFG_CPUSET=0x0005: use core 0 and 2 (Nodes with 2 Hyper-Threading CPUs)
   //   ToDo: define on runtime per taskset + sched_getaffinity ? (jobsystem?)
   // use --byslot for numbits(CFG_CPUSET)<mpi_n
   cpu_set_t cpuset; // typedef struct { ULong bits[CPU_SETSIZE / NCPUBITS]; }
   // new CPU pinning method for more than 32 cores
   struct s_pin {int ncpu;int ofs;int dist;int size;}; // new2013-05-07
   // most important value first (numcpus, then offset, then distance,...)
   int  ncores=0, ii; unsigned long cmask=0L;
   struct s_pin pin = {16,0,1,1}; // CFG_CPUSET=16,0,1,1 mask=0xffff
   // struct s_pin pin = {8,1,4,2}; // mask=0x00066
   for (i1=1;i1<argn;i1++) { /* check all args */  
      if (strncmp(argv[i1],"--pin_ncpu=",11)==0) { 
         pin.ncpu=atoi(argv[i1]+11); continue; }   
      if (strncmp(argv[i1],"--pin_ofs=",10)==0) {  
         pin.ofs= atoi(argv[i1]+10); continue; }   
      if (strncmp(argv[i1],"--pin_dist=",11)==0) { 
         pin.dist=atoi(argv[i1]+11); continue; }   
      if (strncmp(argv[i1],"--pin_size=",11)==0) { 
         pin.size=atoi(argv[i1]+11); continue; }   
   }

   if (pin.ncpu<1) {printf("ERROR: pin.ncpu<1\n");exit(1);}
   if (pin.size<1) {printf("ERROR: pin.size<1\n");exit(1);}
   if (pin.dist<0) {printf("ERROR: pin.dist<1\n");exit(1);}
   if (pin.ofs>=pin.ncpu) {printf("ERROR: pin.shift\n");exit(1);}

   // auto-detect num cores by try+fail sched_affinity
   CPU_ZERO(&cpuset); 
   for(j1=0,i=CPU_SETSIZE;i;i>>=1){
     CPU_SET(j1|i,&cpuset);
     if (!sched_setaffinity(0, sizeof(cpu_set_t), &cpuset)) j1|=i; // no_err!
     CPU_CLR(j1|i,&cpuset);
   }
   if (!mpi_i) printf("# Max-detected-CPU-Cores/Threads= %d\n", j1+1);
   ncores=j1+1;
      
   for (ii=0;ii<8*sizeof(long) && ii<mpi_n;ii++) {  // set cmask
     j1=((ii/pin.size)*pin.dist
        +(ii%pin.size)+pin.ofs)%pin.ncpu;
     if (j1<8*sizeof(long)) cmask |= 1UL<<j1;
   }
   if (!mpi_i)
     printf("# CPU-Pinning: idx=((i/%d)*%d+(i%%%d)+%d)%%%d, mask=0x...%08lx\n",
     pin.size, pin.dist, pin.size, pin.ofs, pin.ncpu,
      cmask);
   if (pin.ncpu>ncores && !mpi_i) printf("WARNING: pin.ncpu>ncores\n");

   // does only cover equidistant distribution but most
   // MASK     
   // 0xffff = {16,1,1,0}
   // 0x6666 = {16,2,4,1}
   // 0x0f0f = {16,4,8,0} = ((ii/4)*8+0)%16
   // 0xf0f0 = {16,4,8,4}
   ii=((mpi_i/pin.size)*pin.dist
      +(mpi_i%pin.size)+pin.ofs)%pin.ncpu;
   // ncores=pin.ncpu;
   if (ii>=CPU_SETSIZE) {printf("ERROR: ii>CPU_SETSIZE, ii=%d\n",ii);exit(1);}
   CPU_ZERO(&cpuset);   // void CPU_ZERO(cpu_set_t *set); bits[idx]|=mask
   CPU_SET(ii,&cpuset); //  ->bits[ii/NCPUBITS] |= (1<<(ii%NCPUBITS))
   
   // The constant CPU_SETSIZE (1024) specifies a value one greater than  the
   //      maximum CPU number that can be stored in a CPU set.
   usleep((mpi_i%16+1)*10000L); // 10ms..170ms, sort outputs  
   if (mpi_i<4 || mpi_i>=mpi_n-2) {
     printf("# set cpu affinity for core(%d)=%d size=%d\n",
              mpi_i, ii, CPU_SETSIZE);
   }
   // no sched_setaffinity on OSF1
   if (sched_setaffinity(0, sizeof(cpu_set_t), &cpuset)) {
     printf("# set cpu affinity for core(%d)=%d size=%d failed\n",
             mpi_i, ii, CPU_SETSIZE);
             usleep(100000L);perror("set_affinity");
   }
   usleep(100000); // 100ms, pseudo sync
  }
  MPI_Barrier(MPI_COMM_WORLD);
#endif

  // -- add a dummy loop here for 12s for a taskset call
  if (!mpi_i) {
    printf("# MPI:");
#if defined(OMPI_MINOR_VERSION)  /* using stringification of macros */
    printf(" OMPI-%d.%d",OMPI_MAJOR_VERSION,OMPI_MINOR_VERSION); 
#ifdef OMPI_RELEASE_VERSION
    printf(".%d",OMPI_RELEASE_VERSION);
#endif
#elif defined(MPICH_VERSION)
    printf(" MPICH-" MPICH_VERSION); 
#elif defined(MPICH2_VERSION)
    printf(" MPICH2-" MPICH2_VERSION); 
#elif defined(MPICH2)    /* sicortex pathscale mpicc-3.2n */
    printf(" MPICH2 (unknown version)"); 
#else
    printf(" non-OMPI/MPICH");
#endif
    printf(" using MPI %d.%d with %d nodes\n", version, subversion, mpi_n);
    printf("# blocksize: 2^%2d = %12lu = %.3e\n", bsize2, bsize, 1.*bsize);
    printf("# mintime[s]:  %5.2f\n", mintime);
    printf("# pairwise:    %2d\n", pairs);
    printf("# rankshift:   %2d\n", rankshift);
    printf("# align:       %2d\n", align);
    
    if (rankshift==0 || abs(rankshift) > mpi_n/2 || mpi_n % (2*abs(rankshift))) {
       if (!mpi_i) printf("ERROR: mpi_n must be multiple of 2*rankshift\n");
       MPI_Abort(MPI_COMM_WORLD,1); };

    printf("# features: check for data errors at end\n");
  }
  if (align) {  /* 64 byte = DMA friendly */
    buf1 = (char *)memalign(64,bsize); /* malloc.h */
    buf2 = (char *)memalign(64,bsize);
  } else {
    buf1 = (char *)malloc(bsize); /* replace by numa_alloc_local */
    buf2 = (char *)malloc(bsize);
  }
  if ((!buf1)     || (!buf2)) fprintf(stderr,"ERROR: malloc failed");
  /* ToDo: get max error per MPI */
  if ((!buf1)     || (!buf2))     MPI_Abort(MPI_COMM_WORLD, 1);

  /* the last state of node_dst + src is the initial state for the run */
  /* touching and writing memory */
  for (i1=0;i1<bsize;i1++) buf1[i1]=buf2[i1]=(char)i1;

  /* eating the memory to see whether or not libmpi eats vmem at runtime */
  /* this would be worse for spinpack, which tries to fill vmem with data */
  /* sicortex: pathcc -lscmpi = MPICH2
   *  1*6*1kB: internal ABORT - process 4: ... pack.c:130: : xmalloc(16384) failed
   *  1*6*4kB=ok
   * 12*4*1kB=ok
   * 12*5*1kB=size2.PMI_Abort Unable to resolve "ssp": Resolver Error 0 + hang
   */
#if 0  /* 2015-03-31 replaced by reduce-free test */
  if (minfree_kB!=~0UL) {  /* 2014-07-27 memeater */
    char *buf[32]; /* max. 4GB */
    unsigned long mem_eaten=0L;
    int i,j;
    if (!mpi_i)
      printf("# minfree_kB:   %7lu (memeater)\n", minfree_kB);
    /* alloc 8MB blocks until minfree is left */
    buf[0]=(char *)malloc(minfree_kB<<10);
    sleep(1); /* be safe minfree is allocated before concurrent tasks eat all */
    if (!buf[0] && (mpi_i&(mpi_i-1))==0)  /* ToDo: transfer error per mpi */
      printf("WARNING: memeater%03d malloc minfree_kB failed\n",mpi_i);
    /* eating the rest of memory */
    for (i=31;i>=10;i--) {  /* 2G 1G 512M 256M ... 1kB */
      buf[i]=(char *)malloc(1<<i); /* needs minimum 32bit int */
      if (buf[i]) { 
        mem_eaten |= 1<<i;
        for (j=0;j<1<<(i-10);j++) buf[i][j<<10]=j&255; /* touching mem */
      }
    }
    free(buf[0]); /* minfree for mpi routines */
    if ((mpi_i&(mpi_i-1))==0) /* ToDo: get min/max of all threads */
      printf("# task%03d mem eaten kB: %7lu\n", mpi_i, mem_eaten>>10);
    if ((mem_eaten>>10)+1==1<<21 && !mpi_i)
      printf("WARNING: memeater failed, please ulimit vmem to max. 2GB\n");
  } /* minfree set */
  MPI_Barrier(MPI_COMM_WORLD);
#endif
  if (minfree_kB!=~0UL) {
    char *b0[128], *b1;
    int size=31; // start with 2GB, go down
    size_t j;
    if (!mpi_i) printf("# Measure minfree (system may OOM-crash):\n");
    b1=malloc(minfree_kB<<10);
    for (i1=0;i1<128;i1++) {
      if(!mpi_i)printf("TestMinfree%03dsz%02d ",(int)i1,size);
      get_minfree(); // 31s
      // strace: mmap 2G 2G brk 2G+8M  1G 1G brk 1G  512M+64K 512M+64K brk 512M+8M+64K 
      // strace: mmap 256M+64k 256M+64K brk 256M+8M+64K  (1malloc=3mmap(+8M+64K))
      // strace: mmap 64M- 32M- 400M+(PROT_NONE) !? unmap64M mmap16M- mmap64M+(PROT_NONE)
      //    mmap 8M- 4M- 1M- 512K- ... 64k- 64k- # woodoo weggelassen
      // ... malloc macht eigenes nicht nachvollziehbares woodoo
      // crashes here! within get_minfree!
      //   get_minfree() "finished freeing, start Allreduce" + crash 2015-04-01
      if (b1) {free(b1);} b0[i1]=0; // kein unmap hier! libc-interne Tabellen!
      if(!mpi_i)printf("TestMinfree%03dsz%02d ",(int)i1,size);
      get_minfree(); // 31s
      test_sendrecv(bsize); // until crash!
      // strace: mprotect(0x5619580000, 65536, PROT_READ|PROT_WRITE) = -1 ENOMEM
      // strace: mmap(NULL, 134217728, PROT_NONE, MAP_PRIVATE|MAP_ANONYMOUS|MAP_NORESERVE, 425197216, 0x1580000) = 0x561c000000
      //                    128MB
      // strace: munmap(0x5620000000, 67108864)          = 0
      //         mmap(NULL, 65536, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, 1024, 0x3f2eadfcf3300000) = -1 ENOMEM
      // ... malloc macht eigenes nicht nachvollziehbares woodoo
      // ergo: lot of mallocs (because in inner loop) is a bad idea! avoid it!
      b1=malloc(minfree_kB<<10);  // 2048kB
        if(!b1) { 
          b1=malloc(minfree_kB<<9);
          if (!b1) printf("ERR malloc(minfree) failed\n");
          free(b1); b1=NULL;
        }
      if (b1) b0[i1]=malloc(1UL<<size);
      if (b1) while(b0[i1]==NULL && size>10) {size--;b0[i1]=malloc(1UL<<size);}
      if (b0[i1])
        for (j=0;j<1UL<<(size-10);j++) b0[i1][j<<10]=j&255; /* touching mem */
    }
    for (i1=0;i1<32;i1++) if(b0[i1]) free(b0[i1]);
  } else
  {
    if (!ata) {
      if (!mpi_i) printf("# Measure speed of MPI_sendrecv transactions:\n");
      // for (i1=bsize;i1>0;i1>>=1) test_sendrecv(i1); // max to min
      for (i1=bsize0;i1<=bsize;i1<<=1)
        test_sendrecv(i1); // min to max (2011-10)
    }
    if ( ata) {
      if (!mpi_i) printf("# Measure speed of MPI_alltoall transactions:\n");
      // for (i1=bsize/mpi_n;i1>0;i1>>=1) test_alltoall(i1);
      for (i1=bsize0/mpi_n;i1<=bsize/mpi_n;i1<<=1)
        test_alltoall(i1);
    }
  }
  
  /* test for correctness of transferred data (bit flips?) */
  /* after lot of loops we dont know about the origin of the error, ...
   * ...  so we just count it and show what kind it is */
  for (i1=0;i1<bsize;i1++) if (buf1[i1]!=(char)i1 || buf2[i1]!=(char)i1)
    {
      if (num_bflips<16)
        printf("ERROR %d: %08lx %02x %02x %02x (bad data)\n",
               mpi_i, i1, buf1[i1], buf2[i1], (char)i1);
      num_bflips++;
    }
  i2= num_bflips;
  MPI_Allreduce(&i2, &num_bflips, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
  if (!mpi_i) printf("# byte-flip errors: %d\n", num_bflips);

  free(buf2);
  free(buf1);
  free(tarray);
  MPI_Finalize();
  return 0;
}

/*

  P4-1.3GHz-linux 15.3us      390MB/s(16M)..840MB/s(64K) (2)
  ES45-Tru64-5.1B  2.6us 5%   820MB/s    (2)
                   2.79us    1130MB/s 2% (4)
      overload-2x 26.0us 4%   7.2MB/s 0.2%(8) SR=Sendrecv AA=Alltoall
  GS320-16:    16 12us 860MB/s(128k) (SR) 156us 880MB/s(128k) (AA)  
     runon 0,1  8 12us 409MB/s(128K) (SR)  82us 373MB/s(128k) # psr=0,1,2,3..3
                4  6us 373MB/s(128K) (SR)  23us 440MB/s(64K) (AA)
                2  5us 260MB/s(128K) (SR)  11us 409MB/s(256K) (AA)
  GS1280-32:    2  3.1us 4%   715MB/s 4%  (slow 32CPUs/RAD)
                8  3.9us 3%  1960MB/s(512kB)
                             1560MB/s(1MB) 3% 
               32  6.4us 4%  3930MB/s(32kB)
                             4040MB/s(64kB,128kB)
                             4160MB/s(256kB)
                             3500MB/s(512kB)
                             2500MB/s(1MB) 3%  
  Altix330 numa_alloc_local MPI_via_loopback ToDo: interconnect-topology=star
     # monitor interconnect-traffic: linkstat, dplace?
     MPT-MPI-1.2 (insmod xpmem) MPI_USE_XPMEM=1 MPI_XPMEM_VERBOSE=1 ?
      2 2.09us 1.8GB/s(1M) 1.1GB/s(256K) 1node  numa_alloc_local
      2 1.97us 8G(1M) 17G 15G 13G 12G 8G 5G 3G 2G 1G 300M ... no div psr1,11
      4        1.6GB/s(1M) 1.6GB/s(256K) 4nodes
      8 3.58us 2.5GB/s(1M) 2.7GB/s(256K) 4nodes
     12 4.05us 3.8GB/s(1M) 3.4GB/s(256K) 6nodes numa_alloc_local
     12 4.05us 4.1GB/s(1M) 34GB/s(128K) 6nodes malloc lo (mem remap?)
     12 4.05us 4.3GB/s(1M) 32GB/s(128K) 6nodes malloc xpmem
      # MPI_Sendrecv                    MPI_Alltoall
      4 3.2us  4.2GB/s(1M) 16GB/s(128K) 12.4us  9GB/s(1M) 20GB/s(256K) # running spinpack8
      4 3.1us  4.6GB/s(1M) 20GB/s(128K) 12.4us 11GB/s(1M) 23GB/s(512K) # stopped spinpack8 (parallel-fluent-speed*3.5 14.7.08)
     12 3.9us  3.4GB/s(1M) 37GB/s(128K) 42.0us  6GB/s(1M) 26GB/s(512K) # stopped spinpack8
  2DualAMD265:    2  23us 4%   660MB/s 4% (last!)   baldur
  4Xeon3GHz_tina  4 3.3us 4%   307MB/s 4%  Dec07 node004
    2*1G via .1G  2 160us 5%    15MB/s 3%  Dec07 node4+node012
                  4 191us 4%    28MB/s 3%  Dec07 node4+node012+...
                  8 206us 4%    63MB/s 3%  Dec07 node4+node012+...
  2*4Xeon_1GbE_quantum_parastation_mpich_2008 (4kB/44us) 
      2  lat=0.76us aBW=1511MB/s(256kB) Oct11 shared_mem cset
      8  lat=0.87us aBW=3038MB/s(256kB) Oct11 shared_mem cset
      2  lat=48.9us aBW= 241MB/s(8MB)   Oct11  2*GbE aBW=235MB/s(512kB) cset
     16  lat=48.8us aBW=1660MB/s(16MB)  Oct11 16*GbE aBW=570MB/s(512kB) cset
     16  lat=43.6us aBW=1670MB/s(16MB)  n*104MB/s Oct11 1015MB/s(32kB) noCPUSET
  sicortex.kautz 958*6: 2014-06 clk=500MHz ToDo (340GB/s 2009) max.6hops
     1*2 *  550MB/s             (2 hops, away and back = blocking) -64% TP
     2*1 * 1.53GB/s =  3.1GB/s 1.9us (128MB) throughput = BW  (3x multi rail)
     2*6 *  218MB/s =  2.6GB/s 2.4us (16M-256MB)   2013-05
    12*1 ...                  throughput reduced
    81*1 *  730MB/s =   59GB/s ...                  -52% blocking net
   243*1 *  500MB/s =  120GB/s 2.6us                -67% blocking net
   958*1 *  316MB/s =  303GB/s 3.7us nur 256 loops! -80% blocking net
   958*6 *   48MB/s =  280GB/s 8.3us nur 256 loops! 2013-05  ca. 100*IB-QDR
   955*1 *  305MB/s =  292GB/s 3.6us 2^12-loops*8MB 2017-08 2bad.nodes+2links
   955*6 *   46MB/s =  263GB/s 8.6us 2^14-loops*1MB 2017-08 2bad.nodes+2links
     # but 1st 955*6 was hanging on start send_recv
     m31n24: busybox nc -v -l -p 6660 > /dev/null # start netcat as receiver
     m31n23: dd if=/dev/zero bs=1024K count=512 | busybox nc -v scx-m31n24 6660
     # m31n23-to-n24: 49.9-50.2MB/s  back: 48.7-50.3MB/s     2017.11
     # m31n23-to-n25: 48.4-49.9MB/s  
     # m31n23-to-n26: 45.8-50.1MB/s  16K=31-51.5MB/s 1K=25-50MB/s
     # 2xnetcat=2*43MB/s, pipe is slow! dd=24GB/s pipe=102MB/s(16K++)
     # scp=5MB/s
     # busybox httpd -v -f -p 8888 
     #  wget -v -O- scx-m31n26:8888/a > /dev/null  # 100MB/1.7s=60MB/s
  sicortex.asgard 12*6 2009 clk=680MHz mpich2 3pair-distances: 1.22,1.24,1.27us
     2*1 * 1.72GB/s =  3.4GB/s 1.3us (64MB) 2009 (speed2014 + 1..3%)
    12*6 *  213MB/s = 15.4GB/s 2.4us (64MB)
     2*1 * 1.52GB/s =  3.0GB/s 1.3us (16M-128M)  2014-06 throughput
     2*6 *  240MB/s =  2.9GB/s 1.9us (64M-128MB)  -0% blocking net
    12*1 * 1.41GB/s = 17.0GB/s 1.3us (16M-128MB)  -7% blocking net
    12*6 *  216MB/s = 15.6GB/s 2.5us (2M-128MB)  -10% blocking net
  sicortex.mkautz(2009) 12*6 2014-06 clk=500MHz max.3hops
     2*1 * 1.45GB/s =  2.9GB/s 1.8us (16M-128MB)      multi-rail-bonus! (-0%)
     2*6 *  220MB/s =  2.6GB/s 2.4us  (1M-128MB, 80s) 5% multi-rail-bonus!
    12*1 * 1.36GB/s = 16.3GB/s 1.8us (16M-256MB, 24s)  -7% blocking net
    12*6 *  206MB/s = 14.8GB/s 3.0us  (4M-128MB, 600s) -7% blocking net
    # may crash on 6*2*256MB=3GB of 3.6GB without any msg!
  t100.2016 3.15GB/s/node im best-case bis 120 nodes, unordnung=1.5GB/s/node
  hpc21.2021 12.3GB/s/wire bis 4*40nodes, dann 3.6GB/s ca. 20x 2.1GB/s (054-253)
     p.e. -N 40 -w cn[081-120],  241-260 =  badNode=cn253 goes slower
    50*4 *  5.2GB/s min=4.7GB/s  max=5.4GB/s pair=i+1 local_only job1431
    50*4 *  3.1GB/s = 12.3GB/s/node = 98 Gbit/s = 98% pair=i+n/2 job1431
   100*2 *  7.1GB/s min=6.9GB/s  max=7.4GB/s pair=+1  local_only job1430
   100*2 *  4.1GB/s min=3.1GB/s  max=6.1GB/s pair=+n/2 job1430
   100*1 *  8.3GB/s min=6.1GB/s  max=12.2GB/s pair=+n/2 job1429 62*max+38*min
     congestions depend from net-topology
     1MB/mean(ti) != mean(1/ti) ??? 2/(4s)=1/2/s ?= (1/1s + 1/3s)/2=2/3/s
    40-port-switch = 20*200Gb-Up + 20*(100+100)Gb-down = 40 Nodes/switch
     TODO list of nodes+ports per switches, iblinkinfo 4x53Gbps or 2x/node
     TODO hpc21_overview.txt
       cn050=dev293 to dev233p80, 
       cn052=dev294 to dev233p79
 hpc21.rack01: cn001-cn036=9*4nodes2X=18p4X + 2*stor-4X-links + 20 uplinks?
       5*dev4,9,10,13 to dev8p21-59:2 4X 19*uplinks-unsortiert?
       ibsw2=5*dev4  to dev8p29,33-39:2 4X
       ibsw1=5*dev13 to dev8p21-27:2,31 4X
       NA  to dev8p81 (40+1p ?? interner port? TODO)
       stor05=dev6  to dev8p61  4X fibre to rack09
    x_ stor04=dev36 to dev8p17  4X fibre to rack09
       cn035=dev227 to dev8p14    cn036=dev155 to dev8p66
    9_ cn033=dev46  to dev8p65    cn034=dev39  to dev8p13  ??? vertauscht?
       ...
    8_ cn029
       ...
    7_ cn025
       cn023=dev152 to dev8p2     cn024=dev177 to dev8p78
    6_ cn021=dev171 to dev8p1     cn022=dev162 to dev8p77
       cn019=dev138 to dev8p3     cn020=dev167 to dev8p80
    5_ cn017=dev219 to dev8p4     cn018=dev217 to dev8p79
       ...
    4_ cn013
       ...
    3_ cn009
       cn007=dev161 to dev8p15    cn008=dev195 to dev8p67
    2_ cn005=dev151 to dev8p16    cn006=dev199 to dev8p68
       cn003=dev33 to dev8p19     cn004=dev34 to dev8p63
    1_ cn001=dev30 to dev8p20     cn002=dev31 to dev8p64
       3*Power/Node-to-1Phase1PDU
  ToDo: show scaling speed(threads)

*/
